source ../library/tkEmacs.tcl

wm maxsize . 1000 1000
wm geometry . 800x600
tkemacs .ff0 -width 100 -height 300
pack append . .ff0 {top}
tkemacs .ff1 -width 300 -height 200
pack append . .ff1 {top}
if {$tkVersion >= 3.3} {
  button .ff2 -text {Quit} -command {exit 0}
} {
  button .ff2 -text {Quit} -command {catch "destroy ."; catch "exit 0"}
}
pack append . .ff2 {bottom}

