source ../library/tkGS.tcl

if {$tkVersion >= 3.3} {
  button .q -text Quit -command {exit 0}
} {
  button .q -text Quit -command {catch "destroy ."; catch "exit 0"}
}
tkgs .f -file test.ps -width 200 -height 200 -bbox {37 652 149 758}
pack append . .f {top fill expand} .q {top}
wm maxsize . 1000 1000

# eof

